import abc
from typing import Any
import json

from .connect import ConnectionABC, TCPConnection


Payload = dict
JsonStr = str
Response = JsonStr


class SerialyzerError(Exception): ...


class Serialyzer:
    serialyzer = json

    @classmethod
    def dump(cls, dict_obj):
        return cls.serialyzer.dumps(dict_obj)

    @classmethod
    def load(cls, json_obj):
        return cls.serialyzer.loads(json_obj)


class ClientABC(abc.ABC):
    @abc.abstractmethod
    def _serialize(self, payload: Payload) -> str: ...
    @abc.abstractmethod
    def _deserialize(self, response: str) -> dict: ...
    @abc.abstractmethod
    def _execute(self, method: str, *args, **kwargs) -> Response: ...
    @abc.abstractmethod
    def _handle_response(self, response: dict) -> Any: ...


class ClientError(Exception): ...


class ClientBase(ClientABC):
    def __init__(self, host, port, connector: ConnectionABC=TCPConnection, 
                 serialyzer: type=Serialyzer, **conn_kwargs):
        self.addr = (host, port)
        self._serialyzer = serialyzer
        self._conn = connector(host, port, **conn_kwargs)
        self._conn.connect()

    def _serialize(self, payload):
        return self._serialyzer.dump(payload)

    def _deserialize(self, response):
        return self._serialyzer.load(response)

    def _execute(self, method, *args, **kwargs):
        payload = {'endpoint': method,
                   'args': args,
                   'kwargs': kwargs}

        self._conn.send(self._serialize(payload))
        return self._conn.recv()

    def _handle_response(self, response):
        if response['type'] == 'error':
            err = response['message']
            raise ClientError(f'Couldn\'t execute: {err}')

        return response['message']

    def getsockname(self):
        return self._conn.getsockname()

    def close(self) -> None:
        self._conn.close()



import unittest

from pyshard.storage import InMemoryStorage
from pyshard.storage.errors import IndexNotFoundError


class TestInMemoryStorage(unittest.TestCase):
    def setUp(self):
        self.storage = InMemoryStorage()

    def _create_index(self, index):
        # check IndexNotFoundError
        with self.assertRaises(IndexNotFoundError) as context:
            self.storage.read(index, 'test')

            self.assertTrue(index in context)

        self.storage.create_index(index)

        # check index is created
        self.assertTrue(index in self.storage.indexes)

    def test_write_and_read(self):
        index = 'test'
        key = 'test_key'
        value = 'test_value'

        self._create_index(index)

        self.storage.write(index, key, value)
        self.assertEqual(self.storage.read(index, key), value)

    def test_read_not_exists(self):
        index = 'test'
        key = 'test_key'

        self._create_index(index)

        self.assertIsNone(self.storage.read(index, key))

    def test_write_to_index_not_exists(self):
        index = 'test'
        key = 'test_key'
        value = 'test_value'

        with self.assertRaises(IndexNotFoundError) as context:
            self.storage.write(index, key, value)

            self.assertTrue(index in context)

    def test_remove_and_read(self):
        index = 'test'
        key = 'test_key'
        value = 'test_value'

        self._create_index(index)

        self.storage.write(index, key, value)
        self.assertEqual(self.storage.pop(index, key), value)
        self.assertIsNone(self.storage.read(index, key))

    def test_pop_and_read(self):
        index = 'test'
        key = 'test_key'
        value = 'test_value'

        self._create_index(index)

        self.storage.write(index, key, value)
        self.storage.pop(index, key)
        self.assertIsNone(self.storage.read(index, key))

    def test_drop_index(self):
        index = 'test'

        self._create_index(index)

        self.storage.drop_index(index)
        self.assertTrue(index not in self.storage.indexes)



import sys
import argparse
import inspect
import json

from pyshard import Pyshard
from pyshard.settings import settings


REGISTRY = dict()
SEPARATOR = '|'


def _register(name):
    def _wrapper(func):
        REGISTRY[name] = func

        return func

    return _wrapper


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('func', type=str)
    parser.add_argument('index', type=str)
    parser.add_argument('-b', '--bootstrap-server', type=str)
    parser.add_argument('--force', action='store_true')

    args = parser.parse_args()

    dispatch_and_execute(args.__dict__)


def dispatch_and_execute(args):
    func = REGISTRY[args.pop('func')]
    args = {key: value for key, value in args.items()
            if key in inspect.getfullargspec(func).args}
    func(**args)


@_register('cat')
def cat(index, bootstrap_server=None):
    bootstrap_server = _retrieve_bootstrap_server(bootstrap_server)

    with Pyshard(bootstrap_server=bootstrap_server) as app:
        for key in app.keys(index):
            doc = app.read(index, key).result
            sys.stdout.write(f'{key}{SEPARATOR}{json.dumps(doc)}\n')
            sys.stdout.flush()


def _process_doc(raw_doc):
    if raw_doc.startswith('{') and raw_doc.endswith('}'):
        doc = json.loads(raw_doc)
    elif raw_doc.isdigit():
        doc = int(raw_doc)
    elif isfloat(raw_doc):
        doc = float(raw_doc)
    else:
        doc = raw_doc

    return doc


def isfloat(raw_doc):
    parts = raw_doc.split('.')
    if len(parts) != 2:
        return False
    if not ((parts[0].isdigit() or not parts[0]) and parts[1].isdigit()):
        return False

    return True


@_register('write')
def write(index, bootstrap_server=None, force=False):
    bootstrap_server = _retrieve_bootstrap_server(bootstrap_server)

    with Pyshard(bootstrap_server=bootstrap_server) as app:
        if force:
            app.create_index(index)

        for line in sys.stdin:
            key, raw_doc = line.rstrip('\n').split(SEPARATOR)
            doc = _process_doc(raw_doc)
            app.write(index, key, doc)


def _retrieve_bootstrap_server(bootstrap_server):
    if bootstrap_server:
        bs = bootstrap_server.split(':')
        return [bs[0], int(bs)]
    else:
        return settings.BOOTSTRAP_SERVER


@_register('test_func')
def test_func(index, bootstrap_server):
    print(index, bootstrap_server)



from collections import defaultdict

from ..storage import InMemoryStorage
from .client import ShardClient
from ..utils import get_size


class Shard:
    def __init__(self, start, end, storage_class=InMemoryStorage, max_size=1024, bins_num=5,
                 buffer_size=1024,
                 **storage_kwargs):
        self._name = None
        self._empty = True
        self.storage = storage_class(**storage_kwargs)
        self.storage.start()
        self._buffer_size = buffer_size

        self.size = 0
        self.max_size = max_size
        self._start = start
        self._end = end
        self._bins_num = bins_num
        self._bin_step = self.estimate_bin_step()
        self._distr = defaultdict(int)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def distr(self):
        return self._distr

    def update_distr(self):
        self._distr = defaultdict(int)
        for doc in self.storage.values():
            hash_ = doc['hash_']
            bin_ = self._get_bin(hash_)
            self._distr[bin_] += 1

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, value):
        assert isinstance(value, float)
        self._bin_step = self.estimate_bin_step()
        self._start = value

    @property
    def end(self):
        return self._end

    @end.setter
    def end(self, value):
        assert isinstance(value, float)
        self._bin_step = self.estimate_bin_step()
        self._end = value

    def estimate_bin_step(self):
        if self._start is not None and self._end is not None:
            bin_step = (self._end - self._start) / self._bins_num
        else:
            bin_step = None

        return bin_step

    def _get_bin(self, hash_):
        steps = (hash_ - self._start) // self._bin_step
        bin_ = self._start + (self._bin_step * steps)

        return bin_

    @property
    def empty(self):
        return self.storage.empty

    @property
    def free_mem(self):
        return self.max_size - self.size

    def write(self, index, key, hash_, record):
        item_size = get_size(record)
        if self.size + item_size > self.max_size:  # TODO replace memory control to storage
            raise MemoryError(f'Wow! Such data! So big!')

        doc = {'hash_': hash_, 'record': record}

        offset = self.storage.write(index, key, doc)
        if offset == 0:  # TODO replace memory control to storage
            return 0

        self.size += item_size

        bin_ = self._get_bin(hash_)
        self.distr[bin_] += 1

        return item_size

    def has(self, index, key):
        return self.storage.has(index, key)

    def read(self, index, key):
        return self.storage.read(index, key)

    def pop(self, index, key):
        doc = self.storage.pop(index, key)
        if doc is None:
            return

        item_size = get_size(doc['record'])
        self.size -= item_size

        bin_ = self._get_bin(doc['hash_'])
        self.distr[bin_] -= 1

        return doc

    def remove(self, index, key):
        doc = self.storage.pop(index, key)
        if doc is None:
            return 0

        item_size = get_size(doc['record'])
        self.size -= item_size

        bin_ = self._get_bin(doc['hash_'])
        self.distr[bin_] -= 1

        return item_size

    def reloc(self, index, key, pipe: ShardClient):
        # relocates item from remote shard
        item = pipe.pop(index, key)

        if item:
            return self.write(index, key, **item)
        else:
            return 0

    def create_index(self, index):
        self.storage.create_index(index)

    def keys(self, index):
        return self.storage.keys(index)

    def get_stat(self):
        stat = {
            'start': self.start,
            'end': self.end,
            'empty': self.empty,
            'max_size': self.max_size,
            'free_mem': self.free_mem,
            'distribution': dict(self.distr)
        }

        return stat

    def close(self):
        self.storage.stop()



import abc
import struct
import logging
import socket
import asyncio

from .typing import Codec

logger = logging.getLogger(__name__)
Kb = 1024


def mksock(host, port, backlog=5, mode='l') -> socket.socket:
    """
    Returns non-blocking socket

    :param host:
    :param port:
    :param backlog:
    :param mode:
    :return:
    """
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.setblocking(False)
    if mode == 'l':
        sock.bind((host, port))
        sock.listen(backlog)

    return sock


class ProtocolABC(abc.ABC):
    @abc.abstractmethod
    def _pack(self, obj: bytes) -> bytes: ...
    @abc.abstractmethod
    def do_send(self, bytes_data: bytes, sock: socket.socket) -> None: ...
    @abc.abstractmethod
    def do_recv(self, sock: socket.socket) -> bytes: ...


class AsyncProtocolABC(abc.ABC):
    @abc.abstractmethod
    def _pack(self, obj: bytes) -> bytes: ...
    @abc.abstractmethod
    async def do_send(self, bytes_data: bytes, conn: socket.socket) -> None: ...
    @abc.abstractmethod
    async def do_recv(self, sock: socket.socket) -> bytes: ...


class ConnectionABC(abc.ABC):
    @abc.abstractmethod
    def connect(self) -> None: ...
    @abc.abstractmethod
    def send(self, str_obj: str) -> None: ...
    @abc.abstractmethod
    def recv(self) -> str: ...


class Protocol(ProtocolABC):
    def __init__(self, buffer_size: int=Kb, codec: Codec='utf-8'):
        self._prefix = struct.Struct('I')
        self._buffer_size = buffer_size
        self._codec = codec

    def _pack(self, obj):
        prefix = self._prefix.pack(len(obj))
        return prefix + obj

    def do_send(self, bytes_data: bytes, sock=None):
        sock = sock or self._sock
        sock.sendall(self._pack(bytes_data))

    def do_recv(self, sock=None):
        data = b''
        total = 0
        sock = sock or self._sock

        prefix = sock.recv(self._prefix.size)
        logger.debug(f"Peer received prefix: {prefix}")
        if not prefix:
            raise RuntimeError('Connection was closed by peer')

        msg_len = self._prefix.unpack(prefix)[0]
        logger.debug(f"Peer will receive message of length {msg_len} bytes")

        while total < msg_len:
            buff_size = min(msg_len - total, self._buffer_size)
            chunk = sock.recv(buff_size)
            if not chunk:
                raise AssertionError(f'Expected {msg_len} bytes, received: {len(data)} bytes')

            data += chunk
            total += buff_size

        return data


class AsyncProtocol(AsyncProtocolABC):
    def __init__(self, buffer_size: int=Kb, loop=None, codec: Codec='utf-8'):
        self._prefix = struct.Struct('I')
        self._buffer_size = buffer_size
        self._codec = codec
        self._loop = loop if loop else asyncio.get_event_loop()

    def _pack(self, obj):
        prefix = self._prefix.pack(len(obj))
        return prefix + obj

    async def do_send(self, bytes_data: bytes, conn):
        await self._loop.sock_sendall(conn, self._pack(bytes_data))

    async def do_recv(self, conn):
        data = b''
        total = 0

        prefix = await self._loop.sock_recv(conn, self._prefix.size)
        logger.debug(f"Peer received prefix: {prefix}")
        if not prefix:
            raise RuntimeError('Connection was closed by peer')
            
        msg_len = self._prefix.unpack(prefix)[0]
        logger.debug(f"Peer will receive message of length {msg_len} bytes")

        while total < msg_len:
            buff_size = min(msg_len - total, self._buffer_size)
            chunk = await self._loop.sock_recv(conn, buff_size)
            if not chunk:
                raise AssertionError(f'Expected {msg_len} bytes, received: {len(data)} bytes')

            data += chunk
            total += buff_size

        return data


def _to_bytes(str_obj: str, codec: Codec) -> bytes:
        return bytes(str_obj, encoding=codec)


def _from_bytes(bytes_obj: bytes, codec: Codec) -> str:
        return bytes_obj.decode(codec)


class ConnectionBase(ConnectionABC, Protocol):
    def __init__(self, host: str=None, port: int=None, **protocol_kwargs):
        if host and port:
            self._addr = (host, port)

        super(ConnectionBase, self).__init__(**protocol_kwargs)

    def connect(self):
        raise NotImplementedError()

    def send(self, str_obj):
        bytes_obj = self._to_bytes(str_obj)

        self.do_send(bytes_obj)

    def _to_bytes(self, str_obj: str) -> bytes:
        return bytes(str_obj, encoding=self._codec)

    def recv(self):
        bytes_obj = self.do_recv()

        return self._from_bytes(bytes_obj)

    def _from_bytes(self, bytes_obj: bytes) -> str:
        return bytes_obj.decode(self._codec)

    def close(self):
        raise NotImplementedError()


class TCPConnection(ConnectionBase):
    def __init__(self, *args, **kwargs):
        self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        super(TCPConnection, self).__init__(*args, **kwargs)

    def connect(self):
        self._sock.connect(self._addr)

    def getsockname(self):
        return self._sock.getsockname()

    def close(self):
        self._sock.close()



from .app.app import Pyshard
from .shard.server import ShardServer
from .master.client import MasterClient
from .master.master import BootstrapServer


__all__ = [
    'Pyshard', 'ShardServer', 'MasterClient', 'BootstrapServer'
]



from socket import socket, AF_INET, SOCK_STREAM
import struct


class Protocol:
	
	def __init__(self, buffer_size=1024):
		self._sock = self._make_sock()
		self._buffer_size = buffer_size
		self._prefix = struct.Struct(f'I')

	@staticmethod
	def _make_sock():
		sock = socket(AF_INET, SOCK_STREAM)

		return sock
	
	def _recvall(self, sock):
		data = b''
		total = 0

		prefix = sock.recv(self._prefix.size)
		if not prefix:
			raise RuntimeError('Connection was closed by peer')

		msg_len = self._prefix.unpack(prefix)[0]
		
		while total < msg_len:
			buff_size = min(msg_len - total, self._buffer_size)
			chunk = sock.recv(buff_size)
			print('chunk recieved:', chunk)

			data += chunk
			total += buff_size

		return self._unpack(data)

	def _unpack(self, bdata):
		data = bdata.decode('UTF-8')
		return data

	def _sendall(self, sock, msg):
		data = self._pack(msg)
		length = len(data)
		prefix = self._prefix.pack(length)
		data = prefix + data
	
		sock.sendall(data)

	def _pack(self, data):
		bdata = bytes(data, encoding='UTF-8')
		return bdata

	def close(self):
		self._sock.close()



import sys
from ..core.typing import Codec


def to_bytes(str_obj: str, codec: Codec) -> bytes:
    return bytes(str_obj, encoding=codec)


def from_bytes(bytes_obj: bytes, codec: Codec) -> str:
    return bytes_obj.decode(codec)


def get_size(obj):
    if isinstance(obj, dict):
        size = sum((get_size(v) for v in obj.values()))
    else:
        size = sys.getsizeof(obj)

    return size



from ..core.typing import (Addr, Key, Hash, Doc, Offset)
from ..core.client import ClientABC, ClientBase


def mkpipe(addr: Addr, **kwargs) -> ClientABC:
    """
    Create instance of Client to connect to another shard

    :param addr: host and port to connect
    :param kwargs: params for Client instance
    :return: Client instance
    """
    return ShardClient(*addr, **kwargs)


class ShardClient(ClientBase):
    def write(self, index, key: Key, hash_: Hash, doc: Doc) -> Offset:
        record = {"record": doc, "hash_": hash_}
        response = self._deserialize(self._execute("write", index, key, **record))

        return self._handle_response(response)

    def has(self, index, key: Key):
        response = self._deserialize(self._execute("has", index, key))

        return self._handle_response(response)

    def read(self, index, key: Key):
        response = self._deserialize(self._execute("read", index, key))

        return self._handle_response(response)

    def pop(self, index, key: Key):
        response = self._deserialize(self._execute("pop", index, key))

        return self._handle_response(response)

    def remove(self, index, key: Key):
        response = self._deserialize(self._execute("remove", index, key))

        return self._handle_response(response)

    def open_pipe(self, host, port):
        response = self._deserialize(self._execute("open_pipe", (host, port)))

        return self._handle_response(response)

    def close_pipe(self):
        response = self._deserialize(self._execute("close_pipe"))

        return self._handle_response(response)

    def reloc(self, index, key, addr):
        response = self._deserialize(self._execute("reloc", index, key, addr))

        return self._handle_response(response)

    def get_stat(self):
        response = self._deserialize(self._execute("get_stat"))

        return self._handle_response(response)

    def lock_shard(self):
        response = self._deserialize(self._execute("lock_shard"))

        return self._handle_response(response)

    def release_shard(self):
        response = self._deserialize(self._execute("release_shard"))

        return self._handle_response(response)

    def change_role(self, role, token=None):
        response = self._deserialize(self._execute("change_role", self.getsockname(), role, token=token))

        return self._handle_response(response)

    def set_start(self, value):
        response = self._deserialize(self._execute("set_start", value))

        return self._handle_response(response)

    def set_end(self, value):
        response = self._deserialize(self._execute("set_end", value))

        return self._handle_response(response)

    def update_distr(self):
        response = self._deserialize(self._execute("update_distr"))

        return self._handle_response(response)

    def create_index(self, index):
        response = self._deserialize(self._execute("create_index", index))

        return self._handle_response(response)

    def keys(self, index):  # TODO: bulk operation
        response = self._deserialize(self._execute("keys", index))

        return self._handle_response(response)

    def set_maxsize(self, size):
        response = self._deserialize(self._execute("set_maxsize", size))

        return self._handle_response(response)

    @property
    def name(self):
        response = self._deserialize(self._execute("get_name"))

        return self._handle_response(response)

    @name.setter
    def name(self, name):
        response = self._deserialize(self._execute("set_name", name))

        self._handle_response(response)



import abc
from typing import Union

from ..master.master import Master, _Shards
from ..master.client import MasterClient
from ..shard.client import ShardClient
from ..core.client import ClientError
from ..core.typing import Key, Doc, Hash


class AbstractResult(abc.ABC):
    @abc.abstractmethod
    def result(self) -> Union[int, Doc]: ...
    @abc.abstractmethod
    def hash(self) -> Hash: ...
    @abc.abstractmethod
    def __iter__(self): ...


class PyshardABC(abc.ABC):
    @abc.abstractmethod
    def write(self, index, key: Key, doc: Doc) -> AbstractResult: ...
    @abc.abstractmethod
    def read(self, index, key: Key) -> AbstractResult: ...
    @abc.abstractmethod
    def pop(self, index, key: Key) -> AbstractResult: ...
    @abc.abstractmethod
    def remove(self, index, key: Key) -> AbstractResult: ...
    @abc.abstractmethod
    def create_index(self, index): ...


def _map_shards(bootstrap_client, **kwargs):
    shard_map = {}
    map_ = bootstrap_client.get_map()
    for bin, addr in map_.items():
        shard_map[float(bin)] = ShardClient(*addr, **kwargs)

    return _Shards(shard_map)


class Result(AbstractResult):
    def __init__(self, result, hash_):
        self._result = result
        self._hash = hash_

    @property
    def result(self):
        return self._result

    @property
    def hash(self):
        return self._hash

    def __iter__(self):
        yield from [self.result, self.hash]


class Pyshard(PyshardABC):
    def __init__(self, bootstrap_server, buffer_size=1024, master_class=Master,
                 **master_args):
        self._bootstrap_client = MasterClient(*bootstrap_server, buffer_size=buffer_size)
        shards = _map_shards(self._bootstrap_client)  # TODO: add ShardClient kwargs
        self._master = master_class(shards=shards, **master_args)

    def write(self, index, key, doc) -> Result:
        hash_, shard = self._master.get_shard(index, key)
        try:
            offset = shard.write(index, key, hash_, doc)
        except ClientError as err:
            # log warning: err
            res = 0
        else:
            res = offset

        return Result(res, hash_)

    def has(self, index, key) -> Result:
        hash_, shard = self._master.get_shard(index, key)

        return Result(shard.has(index, key), hash_)

    def read(self, index, key) -> Result:
        hash_, shard = self._master.get_shard(index, key)
        try:
            doc = shard.read(index, key)
        except ClientError as err:
            # log warning: err
            res = None
        else:
            res = doc

        return Result(res, hash_)
        
    def pop(self, index, key) -> Result:
        hash_, shard = self._master.get_shard(index, key)
        try:
            doc = shard.pop(index, key)
        except ClientError as err:
            # log warning: err
            res = None
        else:
            res = doc

        return Result(res, hash_)

    def remove(self, index, key) -> Result:
        hash_, shard = self._master.get_shard(index, key)
        try:
            offset = shard.remove(index, key)
        except ClientError as err:
            # log warning: err
            res = 0
        else:
            res = offset

        return Result(res, hash_)

    def create_index(self, index):
        self._master.create_index(index)

    def keys(self, index):
        for shard in self._master.shards:
            for key in shard.keys(index):
                yield key

    def close(self):
        self._bootstrap_client.close()
        self._master.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



from typing import *

Key = Union[int, float, str]
Doc = Union[int, float, str, dict, list, tuple]
Hash = float
Offset = int
Addr = Tuple[str, int]

Codec = str

Request = Tuple[str, list, dict]
Response = str
rRequest = str
rResponse = str



class Node:
	def __init__(self, addr):
		self.addr = addr
		self._acquired = False

	def acquire(self):
		self._acquired = True

	@property
	def acquired(self):
		return self._acquired

	def release(self):
		self._acquired = False



class App:
	def __init__(self, master):
		self._master = master

	def write(self, key, item):
		hash_key_, shard = self._master.get_shard(key)
		try:
			size = shard.write(key, hash_key_, item)
		except MemoryError:
			size = 0

		return size

	def read(self, key):
		_, shard = self._master.get_shard(key)

		return shard.read(key)

	def __getattr__(self, attr):
		return getattr(self._master, attr)


import os
import json

from .base import BaseStorage
from .errors import IndexNotFoundError, IndexExistsError


class InMemoryStorage(BaseStorage):
    def __init__(self, dump_filepath=None):
        self._storage = dict()
        self._dump_filepath = dump_filepath

    @property
    def indexes(self):
        return self._storage.keys()

    def has(self, index, key):
        collection = self._get_index(index)
        return key in collection

    def read(self, index, key):
        collection = self._get_index(index)
        return collection.get(key)

    def write(self, index, key, record):
        collection = self._get_index(index)
        if key in collection:
            return 0
        collection[key] = record

    def pop(self, index, key):
        collection = self._get_index(index)
        return collection.pop(key, None)

    def remove(self, index, key):
        collection = self._get_index(index)
        del collection[key]

    def create_index(self, index):
        if index in self._storage:
            raise IndexExistsError(index)
        self._storage[index] = dict()

    def drop_index(self, index):
        del self._storage[index]

    def values(self):
        for index in self.indexes:
            for value in self.index_values(index):
                yield value

    def index_values(self, index):
        collection = self._get_index(index)
        for key in collection:
            yield collection[key]

    @property
    def empty(self):
        for index in self.indexes:
            if self._storage[index]:
                return False
        return True

    def _get_index(self, index):
        if index not in self._storage:
            raise IndexNotFoundError(index)

        return self._storage[index]

    def keys(self, index):
        collection = self._get_index(index)
        return list(collection.keys())

    def start(self):
        if not self._dump_filepath:
            return

        if os.path.exists(self._dump_filepath):
            with open(self._dump_filepath, 'r') as f:
                self._load_dump(f)

    def _load_dump(self, file):
        data = json.load(file)
        self._storage = data

    def stop(self):
        if not self._dump_filepath:
            return

        with open(self._dump_filepath, 'w') as f:
            self._dump(f)

    def _dump(self, file):
        json.dump(self._storage, file)

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()



import json

import logging

from ..settings import settings
from ..core.server import ServerBase
from .shard import Shard
from .client import mkpipe


logger = logging.getLogger(__name__)


class _Server(ServerBase):
    def __init__(self, host, port, buffer_size, loop,
                 serialize=json.dumps, deserialize=json.loads,
                 backlog=5):
        self._shard_locked = False

        super(_Server, self).__init__(host, port, buffer_size, loop, serialize, deserialize, backlog)

    @classmethod
    def with_shard_lock(cls, method):
        async def method_with_lock(self, *args, **kwargs):
            if self._shard_locked:
                raise Exception("Shard is locked")

            return await method(self, *args, **kwargs)

        return method_with_lock


class ShardServer(_Server):
    def __init__(self, host, port, buffer_size=1024, loop=None, **shard_kwargs):
        self._shard = Shard(**shard_kwargs)
        self._pipe = None

        super(ShardServer, self).__init__(host, port, buffer_size, loop)

    @_Server.endpoint('write')
    @_Server.with_shard_lock
    async def write(self, index, key, hash_, record):
        return self._shard.write(index, key, hash_, record)

    @_Server.endpoint('has')
    @_Server.with_shard_lock
    async def has(self, index, key):
        return self._shard.has(index, key)

    @_Server.endpoint('read')
    @_Server.with_shard_lock
    async def read(self, index, key):
        return self._shard.read(index, key)

    @_Server.endpoint('pop')
    @_Server.with_shard_lock
    async def pop(self, index, key):
        return self._shard.pop(index, key)

    @_Server.endpoint('remove')
    @_Server.with_shard_lock
    async def remove(self, index, key):
        return self._shard.remove(index, key)

    @_Server.endpoint('open_pipe')
    @_Server.with_shard_lock
    async def open_pipe(self, *args, **kwargs):
        if self._pipe:
            raise Exception(f'Pipe={self._pipe} already open.')

        self._pipe = mkpipe(*args, **kwargs)

    @_Server.endpoint('close_pipe')
    @_Server.with_shard_lock
    async def close_pipe(self):
        if not self._pipe:
            raise Exception('No working pipe.')

        self._pipe.close()
        self._pipe = None

    @_Server.endpoint('reloc')
    @_Server.with_shard_lock
    async def reloc(self, index, key, addr: list):
        if not self._pipe:
            raise Exception('No working pipe.')
        if self._pipe.addr != tuple(addr):
            raise Exception(f'Wrong pipe. Exists: {self._pipe.addr}, got: {addr}')

        return self._shard.reloc(index, key, self._pipe)

    @_Server.endpoint('get_stat')
    @_Server.with_shard_lock
    async def get_stat(self):
        return self._shard.get_stat()

    @_Server.endpoint('lock_shard', permission_group='master')
    async def lock_shard(self):
        if self._shard_locked:
            raise Exception('Already locked')

        self._shard_locked = True

    @_Server.endpoint('release_shard', permission_group='master')
    async def release_shard(self):
        if not self._shard_locked:
            raise Exception('Shard is not locked')

        self._shard_locked = False

    @_Server.endpoint('change_role')
    @_Server.with_shard_lock
    async def change_role(self, addr, role, token=None):
        addr = tuple(addr)
        if settings.AUTH and not token:
            raise Exception('Token is required')
        print(self.__dict__)
        if role not in self._roles:
            raise Exception(f'Role {role!r} does not exists')
        try:
            chan = self._channels[addr]
        except KeyError:
            raise Exception(f'No such address={addr}')

        chan.permission_group = role

    @_Server.endpoint('set_start', permission_group='master')
    async def set_start(self, value):
        self._shard.start = value

    @_Server.endpoint('set_end', permission_group='master')
    async def set_end(self, value):
        self._shard.end = value

    @_Server.endpoint('update_distr', permission_group='master')
    async def update_distr(self):
        self._shard.update_distr()

    @_Server.endpoint('create_index')
    async def create_index(self, index):
        self._shard.create_index(index)

    @_Server.endpoint('keys')
    async def keys(self, index):
        return self._shard.keys(index)

    @_Server.endpoint('get_name')
    async def get_name(self):
        return self._shard.name

    @_Server.endpoint('set_name', permission_group='master')
    async def set_name(self, name):
        self._shard.name = name

    @_Server.endpoint('set_maxsize', permission_group='master')
    async def set_maxsize(self, size):
        assert size > self._shard.size, f'Can\'t apply {size}b. ' \
                                        f'Current size: {self._shard.size}b'

        self._shard.max_size = size

    def close(self):
        self._shard.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



import sys
import unittest
from unittest.mock import patch
from io import StringIO
import json

from pyshard.console import pyshard as console


class TestConsole(unittest.TestCase):
    TOWRITE = [
        ('2', '{"test": "test"}', json.loads),
        ('3', '42', int),
        ('1', 'test', str),
        ('4', '0.9', float)
    ]

    def _prepare_mock_stdin(self):
        for line in self.TOWRITE:
            sys.stdin.write(f'{line[0]}|{line[1]}\n')
        sys.stdin.flush()
        sys.stdin.seek(0)

    def _read_mock_stdout(self):
        result = []
        sys.stdout.seek(0)
        for line in sys.stdout:
            key, raw_doc = line.rstrip('\n').split('|')
            doc = json.loads(raw_doc)
            result.append((key, doc['record']))

        result.sort(key=lambda x: x[0])

        return result

    @patch('sys.stdin', StringIO())
    @patch('sys.stdout', StringIO())
    def test_write_and_cat(self):
        index = 'test_index'
        self._prepare_mock_stdin()

        console.write(index, force=True)
        console.cat(index)

        self.TOWRITE.sort(key=lambda x: x[0])
        target = [(key, serializer(doc)) for key, doc, serializer in self.TOWRITE]

        self.assertEqual(target, self._read_mock_stdout())



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='pyshard',
    version='0.2.2',
    description='Distributed key-value storage in Python stdlib',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='la9ran9e',
    author_email="tvauritimur@gmail.com",
    url='https://github.com/la9ran9e/pyshard',
    packages=[
        'pyshard',
        'pyshard.app',
        'pyshard.core',
        'pyshard.master',
        'pyshard.shard',
        'pyshard.settings',
        'pyshard.utils',
        'pyshard.storage',
        'pyshard.console'
    ],
    entry_points={
        'console_scripts': [
            'pyshard=pyshard.console.pyshard:main'
        ]
    }
)



from typing import Union, List, Any
from ..core.client import ClientBase


Key = Union[int, float, str]

class MasterClient(ClientBase):
    def __init__(self, *args, **kwargs):
        
        super(MasterClient, self).__init__(*args, **kwargs)

    def get_shard(self, key):
        response = self._deserialize(self._execute("get_shard", key))

        return self._handle_response(response)

    def get_map(self):
        response = self._deserialize(self._execute("get_map"))

        return self._handle_response(response)

    def stat(self):
        response = self._deserialize(self._execute("stat"))

        return self._handle_response(response)

    def create_index(self, index):
        response = self._deserialize(self._execute("create_index", index))

        return self._handle_response(response)



import unittest

from pyshard import Pyshard
from pyshard.utils import get_size
from pyshard.settings import settings


class TestCommands(unittest.TestCase):
    TEST_INDEX = 'test'
    TYPE_CASES = [
        ('test0', 1),
        ('test1', 1.1),
        ('test2', 'test'),
        ('test3', {'test': 'test'}),
        # ('test4', ['test0', 'test1'])
    ]

    @classmethod
    def setUpClass(cls):
        cls.app = Pyshard(bootstrap_server=settings.BOOTSTRAP_SERVER)
        cls.app.create_index(cls.TEST_INDEX)

    def test_types(self):
        for key, doc in self.TYPE_CASES:
            self.assertEqual(self.app.write(self.TEST_INDEX, key, doc).result, get_size(doc),
                             f'couldn\t write key={key}, doc={doc}')
            self.assertEqual(self.app.remove(self.TEST_INDEX, key).result, get_size(doc),
                             f'couldn\t remove key={key}, doc={doc}')

    def test_write_and_read(self):
        key = 'test_key'
        doc = 'test_record'
        self.assertEqual(self.app.write(self.TEST_INDEX, key, doc).result, get_size(doc),
                         f'couldn\t write key={key}, doc={doc}')
        self.assertEqual(self.app.read(self.TEST_INDEX, key).result['record'], doc,
                         f'couldn\t read key={key}, doc={doc}')

        self.assertEqual(self.app.remove(self.TEST_INDEX, key).result, get_size(doc),
                         f'couldn\t remove key={key}, doc={doc}')

    def test_read_not_existing(self):
        key = 'test_key'
        self.app.remove(self.TEST_INDEX, key)
        self.assertEqual(self.app.read(self.TEST_INDEX, key).result, None, f'couldn\t read key={key}')

    def test_write_duplicate(self):
        key = 'test_key'
        doc = 'test_record'
        self.assertEqual(self.app.write(self.TEST_INDEX, key, doc).result, get_size(doc),
                         f'couldn\t write key={key}, doc={doc}')
        self.assertEqual(self.app.write(self.TEST_INDEX, key, doc).result, 0,
                         f'couldn\t write key={key}, doc={doc}')

        self.assertEqual(self.app.remove(self.TEST_INDEX, key).result, get_size(doc),
                         f'couldn\t remove key={key}, doc={doc}')

    def test_write_and_pop(self):
        key = 'test_key'
        doc = 'test_record'
        self.assertEqual(self.app.write(self.TEST_INDEX, key, doc).result, get_size(doc),
                         f'couldn\'t write key={key}, doc={doc}')
        self.assertEqual(self.app.pop(self.TEST_INDEX, key).result['record'], doc,
                         f'couldn\'t populate key={key}, doc={doc}')

    def test_pop_not_existing(self):
        key = 'test_key'
        self.assertEqual(self.app.pop(self.TEST_INDEX, key).result,
                         None, f'couldn\'t populate key={key}')



import hashlib
import bisect
import json
from contextlib import contextmanager

from ..core.server import ServerBase
from ..shard.client import ShardClient


def _normalize_number(num, boundary):
    # Normalizes between 0 and 1
    return float(num % boundary)/boundary


def _hash_key(key, method, boundary):
    hash_function = getattr(hashlib, method)
    hashed_key = int(hash_function(str(key).encode()).hexdigest(), 16)

    return _normalize_number(hashed_key, boundary)


def _make_bins(num):
    bin_step = 1.0/num

    bins = []

    bin_ = 0.0
    step = 0
    while step < num:
        bins.append(bin_)
        bin_ += bin_step
        step += 1

    return bins


import abc
from typing import Union, List, Tuple


Key = Union[int, float, str]
Addr = Tuple[str, int]
Hash = float
Bin = float


def _mark_up(length: int) -> list:
    bin_step = 1.0/length

    bins = []

    bin_ = 0.0
    step = 0
    while step < length:
        bins.append(bin_)
        bin_ += bin_step
        step += 1

    return bins


def _map(class_: type, bins: List[float], addrs: List[Addr], **kwargs):
    return {bin_: class_(*addr, **kwargs) for bin_, addr in zip(bins, addrs)}


class Shards:
    def __init__(self, *addrs, shard_client_class=ShardClient, **shard_kwargs):
        self._bins = _mark_up(len(addrs))
        self._shard_map = _map(shard_client_class, 
                               self._bins, addrs, 
                               **shard_kwargs)
        self._shards = self._shard_map.values()

    @property
    def bins(self):
        return self._bins

    def __getattr__(self, item):
        return getattr(self._shard_map, item)

    def __getitem__(self, item):
        return self._shard_map.__getitem__(item)


class MasterABC(abc.ABC):
    @abc.abstractmethod
    def get_shard(self, index, key: Key) -> Tuple[Hash, ShardClient]: ...
    @abc.abstractmethod
    def create_index(self, index): ...
    @abc.abstractmethod
    def stat(self): ...
    @abc.abstractmethod
    def close(self): ...


class Master(MasterABC):
    def __init__(self, shards: dict, hash_method: str='md5'):
        self._shards = shards
        self._hash_method = hash_method

    @property
    def shards(self):  # TODO: remove values method
        return self._shards.values()
    
    def get_shard(self, index, key):
        key_comp = self._join_key(index, key)
        bin_, hash_ = self._get_bin(key_comp)
        shard = self._shards[bin_]

        return hash_, shard

    def _join_key(self, *parts):
        chain = []
        for part in parts:
            chain.append(str(part))  # TODO: make parts only string

        return ':'.join(chain)

    def _get_bin(self, key: Key) -> Tuple[Bin, Hash]:
        bins = self._shards.bins
        hash_ = _hash_key(key, self._hash_method, 1e7)
        index = bisect.bisect_left(bins, hash_)-1
        bin_ = bins[index]

        return bin_, hash_

    def create_index(self, index):
        # TODO: update meta (for additional shards)
        for shard in self.shards:
            shard.create_index(index)

    def stat(self):
        stat = {}
        for shard in self.shards:
            stat[shard.name] = shard.get_stat()

        return stat

    def close(self):
        self._shards.close()

# class Master:
#   _shard = Shard

#   def __init__(self, *nodes, shards_num=2, method='md5'):
#       self._method = method
#       self._bin_step = 1/shards_num
#       self._bins = _make_bins(shards_num)
#       self._shards = self._make_shards()
#       self._nodes = nodes

#   def _make_shards(self, size=1024):
#       shards = dict()

#       for bin_, node in zip(self._bins, self._nodes):
#           next_bin = bin_+self._bin_step
#           shard = self._shard(*node, start=bin_, end=next_bin, max_size=size)
#           shard.connect()

#           if not shard.ready:
#               shard.init_shard()
#           else:
#               shard.restat()

#           bins[bin_] = shard
        
#       return shards

#   def _get_bin(self, key):
#       id_ = _hash_key(key, self._method, 1e7)
#       i = bisect.bisect_left(self._bins, id_)-1

#       return self._bins[i], id_

#   def get_shard(self, key):
#       bin_, hash_key_ = self._get_bin(key)
#       return hash_key_, self._shards[bin_]

#   def split(self, increase_num, size=1024):
#       num = len(self._shards)+increase_num
#       self._bins = _make_bins(num)
#       old_shards = self._shards
#       self._bin_step = 1/num
#       self._shards = self._make_shards(size=size)
#       for shard in old_shards.values():
#           self._do_distr(shard)

#   def _do_distr(self, shard, tg_shard=None):
#       if not tg_shard:
#           for key, value in shard.values():
#               id_ = value['hash']
#               i = bisect.bisect_left(self._bins, id_)-1

#               bin_ = self._bins[i]
#               tg_shard = self._shards[bin_]

#               tg_shard.reloc(key, shard.node)

#       else:
#           for key in shard.key():
#               tg_shard.reloc(key, shard.node)

#   def insert(self, bin_, node=None, size=1024):
#       if not node:
#           node = self._nodes.getfree()
#           node.acquire()

#       i = bisect.bisect_right(self._bins, bin_)
#       left_bin = self._bins[i-1]
#       right_bin = self._bins[i+1]
#       self._bins[i:i] = [bin_]
#       new_shard = self._shard(node, start=bin_, end=right_bin, max_size=size)
#       new_shard.connect()
#       left_shard = self._shards[left_bin]

#       _to_move = [key for key, value in left_shard.items() if value['hash'] >= bin_]

#       for key in _to_move:
#           new_shard.reloc(key, left_shard.node)

#       self._shards[bin_] = new_shard

#   def remove(self, bin_):
#       i = self._bins.index(bin_)
#       del self._bins[i]
#       tmp = self._shards[bin_]
#       self._shards[bin_]
#       left_bin = self._bins[i-1]
#       tg_shard = self._shards[left_bin]
#       self._do_distr(tmp, tg_shard)

#       tmp.close()
#       tmp.node.release()

#   def rebalance(self, shard):
#       ...

#   @property
#   def stat(self):
#       stat = dict()

#       for shard_id, shard in sorted(self._shards.items()):
#           stat[shard_id] = {
#               'total memory': shard.max_size,
#               'free memory': shard.free_mem,
#               'distribution': dict(shard.distr)

#           }

#       return stat


def _bootstrap(conf_path=None, *args, **kwargs):
    master_token = kwargs.pop('token', None)
    config = _get_config(conf_path)
    shards = _mkshards(config['shards'], *args, **kwargs)
    shards.get_master_role(master_token)
    with shards.lock():
        _mark_shards(shards, config['shards'])
        _config_shards(shards, config['shards'])

    return shards


def _get_config(conf_path=None):
    if conf_path:
        config_file = open(conf_path, mode='r')
    else:
        raise NotImplementedError()

    config = json.load(config_file)

    shards = config['shards']

    if _is_named(shards):
        _check_names(shards)
    else:
        raise NotImplementedError()

    if _is_marked(shards):
        _check_markers(shards)

    return config


def _is_marked(shards):
    l = len(shards)
    shards = [shard for shard in shards if shard.get('start') is not None
              and shard.get('end') is not None]
    if len(shards) < l:
        raise Exception("All or no one shard must be marked")
    if not shards:
        return False

    return True


def _check_markers(shards):
    shards.sort(key=lambda x: x['start'])
    mem_end = 0.0
    for shard in shards:
        start = shard['start']
        end = shard['end']
        if end <= start:
            raise Exception("Shard end must be greater than it\'s start")
        if mem_end != start:
            raise Exception("Shard start must equal to previous shard end or 0.0")
        mem_end = end


def _is_named(shards):
    names_c = 0
    for shard in shards:
        if 'name' in shard:
            names_c += 1

    if names_c == 0:
        return False

    assert names_c == len(shards), "Named selectively"

    return True


def _check_names(shards):
    names = set()
    for shard in shards:
        name = shard['name']
        if name in names:
            raise KeyError(f'Name {name!r} already reserved')
        names.add(name)


class _Shards(dict):
    def __init__(self, *args, **kwargs):
        super(_Shards, self).__init__(*args, **kwargs)
        self._bins = sorted(self.keys())

    @property
    def bins(self):
        return self._bins

    def get_master_role(self, token=None):
        for shard in self.values():
            shard.change_role('master', token)

    @contextmanager
    def lock(self):
        for shard in self.values():
            shard.lock_shard()
        try:
            yield
        finally:
            for shard in self.values():
                shard.release_shard()

    def close(self):
        for shard in self.values():
            shard.close()

    def __setitem__(self, key, value):
        super(_Shards, self).__setitem__(key, value)
        bisect.insort_right(self._bins, key)


def _mkshards(shards_conf, *args, **kwargs):
    shards = _Shards()
    for shard in shards_conf:
        start = shard['start']
        host, port = shard['host'], shard['port']
        shards[start] = ShardClient(host, port, *args, **kwargs)

    return shards


def _mark_shards(shards, shards_conf):
    for shard, shard_conf in zip(shards.values(), shards_conf):  # TODO: remove values method
        start, end = shard_conf['start'], shard_conf['end']
        shard.set_start(start)
        shard.set_end(end)

        shard.update_distr()


def _config_shards(shards, shards_conf):
    for shard, shard_conf in zip(shards.values(), shards_conf):
        size = shard_conf.get('size', None)
        if size:
            shard.set_maxsize(size)

        name = shard_conf['name']
        shard.name = name


class _Server(ServerBase): ...


class BootstrapServer(_Server):
    def __init__(self, *args, config_path=None, master=Master, hash_method='md5',
                 **kwargs):  # TODO: add bootstrap options
        self._shards = _bootstrap(config_path)
        self._master = master(shards=self._shards, hash_method=hash_method)

        super(BootstrapServer, self).__init__(*args, **kwargs)

    @_Server.endpoint('get_map')
    async def get_map(self):
        return {bin_: shard.addr for bin_, shard in self._shards.items()}

    @_Server.endpoint('get_shard')
    async def get_shard(self, index, key):
        hash_, shard = self._master.get_shard(index, key)
        return hash_, shard.addr

    @_Server.endpoint('create_index')
    async def create_index(self, index):
        return self._master.create_index(index)

    @_Server.endpoint('stat')
    async def stat(self):
        return self._master.stat()


import json
import asyncio
from collections import defaultdict
from struct import error as struct_error

import logging

from ..core.typing import Request, Response, rRequest, rResponse
from ..utils import to_bytes, from_bytes
from ..settings import settings

from .connect import AsyncProtocol, mksock


logger = logging.getLogger(__name__)

Kb = 1024

class _Channel(AsyncProtocol):
    def __init__(self, sock, loop, buffer_size=1024):
        self._sock = sock
        self._loop = loop
        self._chan = None
        self.token = None
        self.permission_group = None

        super(_Channel, self).__init__(buffer_size, loop)

    async def connect(self):
        self._chan = await self._loop.sock_accept(self._sock)
        return self

    async def retrieve_token(self):
        rdata = await self.do_recv(self.sock)
        self.token = from_bytes(rdata, self._codec)

    @property
    def sock(self):
        return self._chan[0]

    @property
    def addr(self):
        return self._chan[1]

    async def msg_iterator(self):
        while True:
            try:
                data = await self.do_recv(self.sock)
            except struct_error as err:
                logger.error(f'Couldn\'t unpack message: {err}')
            except RuntimeError as err:
                logger.warning(f'Addr={self.addr}: {err}')
                break
            except AssertionError as err:
                logger.warning(f'Addr={self.addr} send not enought data. {err}')
                break
            else:
                logger.debug(f'Received message from addr={self.addr}: {data}')
                yield from_bytes(data, self._codec)


class Endpoint:
    def __init__(self, path, method, permission_group):
        self._path = path
        self._method = method
        self._permission_group = permission_group

    @property
    def path(self):
        return self._path

    @property
    def method(self):
        return self._method

    @property
    def permission_group(self):
        return self._permission_group


def _auth(func):
    async def wrapper(self, *args, **kwargs):
        chan = await func(self, *args, **kwargs)
        await chan.retrieve_token()
        if chan.token in self._token_storage:
            chan.permission_group = self._token_storage[chan.token].get('group')
            return chan
        else:
            raise KeyError(f'Not authorized user token={chan.token}')

    return wrapper if settings.AUTH else func


class ServerBase(AsyncProtocol):
    def __init__(self, host, port, buffer_size, loop,
                 serialize=json.dumps, deserialize=json.loads,
                 backlog=5):
        self.sock = mksock(host, port, backlog=backlog, mode='l')
        self._default_queue = asyncio.Queue(maxsize=buffer_size)
        self._master_queue = asyncio.Queue(maxsize=buffer_size // 2)
        self._master_group = 'master'
        self._token_storage = defaultdict(dict)
        self._channels = dict()

        self._proc_locker = asyncio.Lock()

        self._serialize = serialize
        self._deserialize = deserialize

        self._routes = dict()
        self._roles = set()
        self._permissions = defaultdict(set)

        self._discover_endpoints()

        super(ServerBase, self).__init__(buffer_size, loop)

    async def _do_run(self):
        await asyncio.gather(self._worker(self._master_queue),
                             self._worker(self._default_queue),
                             self._main_loop())

    def _dispatch(self, endpoint):
        return self._routes[endpoint]

    def _discover_endpoints(self):
        for key in dir(self):
            attr = getattr(self, key)
            if isinstance(attr, Endpoint):
                self._init_endpoint(attr)

    def _init_endpoint(self, endpoint):
        if endpoint.permission_group:
            self._roles.add(endpoint.permission_group)
            self._permissions[endpoint.path].add(endpoint.permission_group)

        self._routes[endpoint.path] = endpoint.method

    @classmethod
    def endpoint(cls, path, permission_group=None):
        def _wrapper(method):
            return Endpoint(path, method, permission_group)

        return _wrapper

    def _check_permission(self, chan, endpoint):
        if not self._permissions[endpoint]:
            return

        if chan.permission_group not in self._permissions[endpoint]:
            raise Exception("Permission denied")

    async def _dispatch_and_execute(self, chan, endpoint, *args, **kwargs):
        self._check_permission(chan, endpoint)
        async with self._proc_locker:
            return await self._dispatch(endpoint)(self, *args, **kwargs)

    async def _worker(self, queue):
        while True:
            chan, endpoint, args, kwargs = await queue.get()
            try:
                rresp = await self._dispatch_and_execute(chan, endpoint, *args, **kwargs)
            except Exception as err:
                resp = self._handle_error_resp(err)
            else:
                resp = self._handle_success_resp(rresp)

            await self.do_send(to_bytes(resp, self._codec), chan.sock)

            queue.task_done()

    async def _main_loop(self):
        async for chan in self._channel_iterator():
            logger.debug(f"Connection accepted from: {chan.addr}")
            self._loop.create_task(self._handle_channel(chan))

        self.sock.close()

    async def _handle_channel(self, chan):
        async for msg in chan.msg_iterator():
            try:
                endpoint, args, kwargs = self._parse_request(msg)
            except Exception as err:
                logger.warning(f"Couldn\'t parse message={msg!r}, addr={chan.addr} error: {err}")
                break

            if chan.permission_group == self._master_group:
                queue = self._master_queue
            else:
                queue = self._default_queue

            await queue.put((chan, endpoint, args, kwargs))

        chan.sock.close()
        del self._channels[chan.addr]

    async def _channel_iterator(self):
        while True:
            logger.debug("Waiting for connection...")
            try:
                yield await self._accept(self.sock, self._loop)
            except KeyError as err:
                logger.warning(err)
            except asyncio.TimeoutError:
                logger.warning('Couldn\'t recieve token from peer')

    @_auth
    async def _accept(self, sock, loop):
        chan = await _Channel(sock, loop).connect()
        self._channels[chan.addr] = chan
        return chan

    def _parse_request(self, request: rRequest) -> Request:
        req = self._deserialize(request)

        return req['endpoint'], req.get('args', list()), req.get('kwargs', dict())

    def _handle_success_resp(self, rresp: rResponse) -> Response:
        resp = {"type": "success", "message": rresp}

        return self._serialize(resp)

    def _handle_error_resp(self, err: Exception) -> str:
        resp = {"type": "error", "message": err.args}

        return self._serialize(resp)



class BaseStorage:
    def read(self, index, key): ...
    def write(self, index, key, record): ...
    def pop(self, index, key): ...
    def remove(self, index, key): ...

    def create_index(self, index): ...
    def drop_index(self, index): ...

    def values(self): ...
    def index_values(self, index): ...

    def empty(self): ...

    def _get_index(self, index): ...

    def keys(self, index): ...
